"""
Module to handle document chapters
"""

from pathlib import Path
from typing import Any, Callable

from pydantic import BaseModel

from icreports.tags import Taggable
from icreports.document import section
from .section import Section


class Chapter(BaseModel, Taggable):
    """
    Class representing a document chapter following the Jupyter Book
    representation. It can be composed of document Sections.

    The Chapter can have an optional root or base section and
    can in addition be broken down into a collection of Sections.

    :param Section base: An optional root or base section
    :param list[Section] sections: A list of Sections in the Chapter
    """

    base: Section | None = None
    sections: list[Section] = []

    @property
    def files(self) -> list[Path]:
        """
        A list of file paths for all sections
        """
        files = []
        if self.base:
            files.append(self.base.file)
        files.extend([s.file for s in self.sections])
        return files

    def has_tag(self, tag: str) -> bool:
        """
        True if any Section has the provided tag
        """
        if self.base and self.base.has_tag(tag):
            return True
        return any(s.has_tag(tag) for s in self.sections)


def filter_children(chapter: Chapter, pred: Callable[[Any], bool]) -> Chapter:
    """
    Return a Chapter with only Sections matching the predicate.

    :param Chapter chapter: The chapter to filter
    :param pred Callable: The predicate to filter on
    """

    base: Section | None = None
    if chapter.base and pred(chapter.base):
        base = chapter.base

    return Chapter(
        base=base,
        sections=[s for s in chapter.sections if pred(s)],
    )


def load_content(root: Path, chapter: Chapter) -> Chapter:
    """
    Load file content into the Chapter

    :param Path root: The absolute path to the top-level source dir
    :param Chapter chapter: The Chapter to load content for
    """

    base: Section | None = None
    if chapter.base:
        base = section.load_content(root, chapter.base)

    return Chapter(
        base=base,
        sections=[section.load_content(root, s) for s in chapter.sections],
    )



import subprocess


class SpellChecker:

    def __init__(self):
        self.dictionary = "en_GB"

    def check(self, doc: str) -> list[str]:
        cmd = f"hunspell -d {self.dictionary} -H << {doc}"
        ret = subprocess.run(
            cmd, check=True, shell=True, capture_output=True, text=True
        )
        return ret.stdout.splitlines()



import shutil

from iccore.test_utils import get_test_data_dir, get_test_output_dir

from icreports.document.jupyter_book import toc as jb_toc
from icreports.document import config, toc
from icreports.rendering.context import RenderContext
from icreports.rendering.book import render


def test_book_publish():
    content_root = get_test_data_dir() / "mock_document"

    content = toc.load_content(content_root,
                               jb_toc.from_jupyterbook(
                                   jb_toc.read(content_root / "_toc.yml")))
    
    build_dir = get_test_output_dir()
    ctx = RenderContext(source_dir = content_root,
                        build_dir = build_dir,
                        config = config.read(content_root / "_config.yml"))

    render(content, ctx)

    shutil.rmtree(build_dir)



"""
Module representing document sections
"""

from pathlib import Path

from pydantic import BaseModel, field_serializer

from icreports.tags import Taggable
import icreports.content
from icreports.content import Content


class Section(BaseModel, Taggable):
    """
    A Section is the smallest piece of a document and
    can be associated with an individual file with
    document content.

    :param Path file: The path to the file with content
    :param Content content: A representation of the file's content
    """

    file: Path
    content: Content | None = None

    @field_serializer("file")
    def serialize_file(self, file: Path, _info):
        return str(file)

    def has_tag(self, tag: str) -> bool:
        """
        True if the provided tag is found in the file content

        :param str tag: The tag to search for
        """
        if not self.content:
            return False
        return self.content.has_tag(tag)


def load_content(root: Path, section: Section) -> Section:
    """
    Load the section's file content

    :param Path root: The absolute path to the top-level source dir
    :param Section section: The section to load the content for
    """
    return Section(
        file=section.file, content=icreports.content.load(root / section.file)
    )



from pathlib import Path

from pydantic import BaseModel


class Hyperlink(BaseModel):

    link: str

    def resolve(self, root: Path) -> str:
        resolved = self.link
        if resolved.startswith("/"):
            resolved = resolved.lstrip("/")
        elif resolved.startswith("."):
            resolved = str(root) + resolved[1:]
        else:
            resolved = str(root) + "/" + resolved
        return resolved

    def is_local(self) -> bool:
        if not self.link.endswith(".md"):
            return False
        return not (self.link.startswith("http") or self.link.startswith("#"))

    def wikify(self) -> str:
        if self.is_local() and not self.link.endswith("index.md"):
            return self.link.split("/")[-1]
        return self.link



"""
Module to handle a document's table of contents
"""

from pathlib import Path
import typing
from typing import Any, Callable

from pydantic import BaseModel

from iccore.serialization import read_model_yaml, write_model_yaml

from icreports.document import part, section

from .part import Part
from .section import Section


class TableOfContents(BaseModel):
    """
    A document's contents based on the Jupyter Book data
    model as per their yaml io format.

    :param str jb_format: A format hint for Jupyter Book
    :param Section root: Path to the root file, such as frontmatter
    :param list[Part] parts: A list of document parts
    """

    format: str = "jb-book"
    root: Section
    parts: list[Part] = []

    @property
    def files(self) -> list[Path]:
        """
        Return paths to all content files in the document
        """
        files = []
        if self.root:
            files.append(self.root.file)
        for p in self.parts:
            files.extend(p.files)
        return files


def read(path: Path) -> TableOfContents:
    """
    Read the table of contents from a yaml file

    :param Path path: Path to the table of contents file
    """
    return typing.cast(TableOfContents, read_model_yaml(path, TableOfContents))


def write(path: Path, toc: TableOfContents) -> None:
    """
    Write the table of contents to file as yaml

    :param TableOfContents toc: The table of contents to write
    :param Path path: The path to the output file
    """
    write_model_yaml(path, toc)


def load_content(root: Path, toc: TableOfContents) -> TableOfContents:
    """
    Load the file content for all elements of the table of content
    """
    return TableOfContents(
        root=section.load_content(root, toc.root),
        format=toc.format,
        parts=[part.load_content(root, p) for p in toc.parts],
    )


def filter_children(
    toc: TableOfContents, pred: Callable[[Any], bool]
) -> TableOfContents:
    """
    Return a table of contents with Parts matching the supplied tag
    """
    return TableOfContents(
        root=toc.root,
        format=toc.format,
        parts=[part.filter_children(p, pred) for p in toc.parts if pred(p)],
    )



"""
A Part of a document, particularly a book. It is a collection
of Chapters.
"""

from pydantic import BaseModel

from icreports.document import part
from icreports.document.jupyter_book import chapter as jb_chapter

from .chapter import Chapter


class Part(BaseModel):
    """
    A Part is an element of a document, particularly a book,
    that is made of a collection of Chapters.

    :param str caption: A caption or title for the Part
    :param list[Chapter] chapters: A list of the Part's Chapters
    """

    caption: str = ""
    chapters: list[Chapter] = []


def from_jupyterbook(jb_part: Part) -> part.Part:
    """
    Convert a part in Jupyter Book format to the internal format
    """
    return part.Part(
        caption=jb_part.caption,
        chapters=[jb_chapter.from_jupyterbook(c) for c in jb_part.chapters],
    )


def to_jupyterbook(ic_part: part.Part) -> Part:
    """
    Convert a part in the internal format to Jupyter Book format
    """

    return Part(
        caption=ic_part.caption,
        chapters=[jb_chapter.to_jupyterbook(c) for c in ic_part.chapters],
    )



"""
Module to handle a document's table of contents
"""

from pathlib import Path
import logging
import typing

from pydantic import BaseModel, field_serializer

from iccore.serialization import read_model_yaml, write_yaml

from icreports.document import toc, section
from icreports.document.jupyter_book import part as jb_part

from .part import Part

logger = logging.getLogger(__name__)


class TableOfContents(BaseModel):
    """
    A document's contents based on the Jupyter Book data
    model as per their yaml io format.

    :param str format: A format hint for Jupyter Book
    :param Path root: Path to the root file, such as frontmatter
    :param list[Part] parts: A list of document parts
    """

    format: str = "jb-book"
    root: Path
    parts: list[Part] = []

    @field_serializer("root")
    def serialize_root(self, root: Path, _info):
        return str(root)


def from_jupyterbook(jb_toc: TableOfContents) -> toc.TableOfContents:
    """
    Convert a Table of Contents in Jupyter Book format to the internal format
    """
    return toc.TableOfContents(
        format=jb_toc.format,
        root=section.Section(file=jb_toc.root),
        parts=[jb_part.from_jupyterbook(p) for p in jb_toc.parts],
    )


def to_jupyterbook(ic_toc: toc.TableOfContents) -> TableOfContents:
    """
    Convert a Table of Contents in the internal format to Jupyter Book format
    """

    return TableOfContents(
        format=ic_toc.format,
        root=ic_toc.root.file,
        parts=[jb_part.to_jupyterbook(p) for p in ic_toc.parts],
    )


def read(path: Path) -> TableOfContents:
    """
    Read the table of contents from a yaml file

    :param Path path: Path to the table of contents file
    """
    return typing.cast(TableOfContents, read_model_yaml(path, TableOfContents))


def write(path: Path, contents: TableOfContents) -> None:
    """
    Write the table of contents to file as yaml

    :param TableOfContents toc: The table of contents to write
    :param Path path: The path to the output file
    """
    write_yaml(path, contents.model_dump())



"""
Module to support rendering of a book to disk given a book description
and a render context
"""

import logging
import shutil
import subprocess
from pathlib import Path

import iccore.filesystem as fs

from icreports.tags import TagFilter
from icreports.document.jupyter_book import toc as jb_toc
from icreports.document import toc, config
from icreports.rendering import document

from .context import RenderContext


logger = logging.getLogger(__name__)


def _copy_sources(
    contents: toc.TableOfContents, ctx: RenderContext, build: config.Build
):
    """
    Copy source files, filtered for the specific build, to the build
    directory.

    Also copy the filtered table of contents
    """

    build_dir = ctx.build_dir / build.name
    fs.copy_file(ctx.source_dir / "_config.yml", build_dir)

    pred = TagFilter(includes=build.include_tags, excludes=build.exclude_tags)
    contents = toc.filter_children(contents, lambda x: pred(x))
    jb_contents = jb_toc.to_jupyterbook(contents)
    jb_toc.write(build_dir / "_toc.yml", jb_contents)

    files_with_ext = [Path(f"{f}.md") for f in contents.files]
    fs.copy_files_relative(files_with_ext, ctx.source_dir, build_dir)


def _jb_render(ctx: RenderContext, builder_args: str):
    args = ""
    if ctx.rebuild:
        args = "--all "

    cmd = f"jb build {args}{ctx.source_dir} --builder {builder_args}"

    with open(ctx.build_dir / "jb_logs.txt", "w", encoding="utf-8") as f:
        subprocess.run(cmd, shell=True, check=True, stdout=f, stderr=f)


def _render_html(doc_name: str, ctx: RenderContext):

    _jb_render(ctx, "html")
    fs.make_archive(
        ctx.build_dir / f"{doc_name}_html", "zip", ctx.build_dir / "html", rename=True
    )


def _render_pdf(doc_name: str, ctx: RenderContext):

    _jb_render(ctx, "pdflatex")
    shutil.move(ctx.build_dir / "latex/book.pdf", ctx.build_dir / f"{doc_name}.pdf")


def render(contents: toc.TableOfContents, ctx: RenderContext):
    """
    Render a book to disk for the given context
    :param TableOfContents contents: the content to render
    :param BuildContext ctx: the context (or settings) to render the book
    """

    contents = toc.load_content(ctx.source_dir, contents)
    document.render(
        contents, ctx, _copy_sources, {"pdf": _render_pdf, "html": _render_html}
    )



"""
Module for rendering media such as images and converting formats
"""

import logging
from pathlib import Path

from icplot import tex
from icplot.tex import TexBuildSettings
from icplot.image_utils import svg_to_png, svg_to_pdf

logger = logging.getLogger(__name__)


def convert(source_dir: Path, output_dir: Path, build_dir: Path | None = None):
    """
    Convert supported media in the source directory to suitable formats
    and save them in the output directory. Optionally use a separate
    build directory for intermediate or temporary file generation.

    :param Path source_dir: The directory to search for source files
    :param Path output_dir: The directory to put converted files in
    :param Path build_dir: Optional directory for intermediate files
    """

    if not source_dir.exists():
        return

    if not build_dir:
        build_dir = output_dir

    logger.info("Start converting media")

    tex_config = TexBuildSettings(
        source=source_dir, build_dir=build_dir, output_dir=output_dir
    )
    tex.build(tex_config)

    svg_files = list(source_dir.glob("*.svg"))
    logger.info("Found %d svg files in %s", len(svg_files), source_dir)
    for svg_file in svg_files:
        svg_to_png(svg_file, output_dir / f"{svg_file.stem}.png")
        svg_to_pdf(svg_file, output_dir / f"{svg_file.stem}.pdf")

    logger.info("Finished converting media")



"""
Module for document configuration
"""

from pathlib import Path

from pydantic import BaseModel, field_serializer

from iccore.serialization import read_yaml, write_yaml

_DEFAULT_KEY = "icreports"


class Build(BaseModel, frozen=True):
    """
    A build configuration for the document

    :param str name: The name of the build
    :param set[str] outputs: Requested output formats: 'src', 'html', 'pdf'
    :param set[str] include_tags: Include only documents with these tags
    """

    name: str
    outputs: set[str] = set()
    include_tags: set[str] = set()
    exclude_tags: set[str] = set()


class Config(BaseModel, frozen=True):
    """
    The document configuation

    :param dict data: The raw data from the config file as dict
    :param str version: The document version
    :param str project_name: The project's name
    :param Path media_dir: Relative location of media content
    :param Path source_dir: Relative location of source files
    :param tuple[Build] builds: Build configurations
    """

    version: str = "0.0.0"
    project_name: str = "document"
    media_dir: Path = Path("src/media")
    source_dir: Path = Path("src")
    builds: list[Build] = []

    @field_serializer("media_dir")
    def serialize_media_dir(self, media_dir: Path, _info):
        return str(media_dir)

    @field_serializer("source_dir")
    def serialize_source_dir(self, source_dir: Path, _info):
        return str(source_dir)


def read(path: Path, key: str = _DEFAULT_KEY) -> Config:
    """
    Read the config from file. The Jupyter Book yaml format
    is assumed, with the config for this project under a
    specified key.


    :param Path path: The path to the config file
    :param str key: The key in the yaml to read the config from
    """
    config = read_yaml(path)
    return Config(**config[key])


def write(path: Path, config: dict, doc_config: Config, key: str = _DEFAULT_KEY):
    """
    Write the config to file, inserting the DocumentConfig under
    a specific key at the same time.

    :param Path path: The path to write to
    :param dict config: The full config in Jupyter Book format
    :param DocumentConfig doc_config: The config for this project
    :param str key: The key to insert the doc_config under
    """
    config[key] = doc_config.model_dump()
    write_yaml(path, config)



from iccore.serialization import read_yaml
from iccore.test_utils import get_test_data_dir, get_test_output_dir

from icreports.document import config


def test_config():

    config_file = get_test_data_dir() / "mock_document/_config.yml"

    original_config = read_yaml(config_file)
    doc_config = config.read(config_file)

    config_out = get_test_output_dir() / "config_out.yml"
    config.write(config_out, original_config, doc_config)

    config1 = config.read(config_out)

    assert doc_config.project_name == config1.project_name

    config_out.unlink()



from pathlib import Path

from pydantic import BaseModel, field_serializer

from icreports.document import section


class Section(BaseModel):
    """
    A Section is the smallest piece of a document and
    can be associated with an individual file with
    document content.

    :param Path file: The path to the file with content
    """

    file: Path

    @field_serializer("file")
    def serialize_file(self, file: Path, _info):
        return str(file)


def from_jupyterbook(jb_section: Section) -> section.Section:
    return section.Section(file=jb_section.file)


def to_jupyterbook(ic_section: section.Section) -> Section:
    return Section(file=ic_section.file)



from pathlib import Path

from pydantic import BaseModel, model_serializer

from icreports.document import section, chapter

from icreports.document.jupyter_book import section as jb_section
from .section import Section


class Chapter(BaseModel):
    """
    Class representing a document chapter following the Jupyter Book
    representation. It can be composed of document Sections.

    The Chapter can have an optional root or base section and
    can in addition be broken down into a collection of Sections.

    :param Path file: The chapter can have a 'root' section here
    :param list[Section] sections: A list of Sections in the Chapter
    """

    file: Path | None = None
    sections: list[Section] = []

    @model_serializer
    def serialize_model(self):
        ret = {}
        if self.file:
            ret["file"] = str(self.file)
        if self.sections:
            ret["sections"] = self.sections
        return ret


def from_jupyterbook(jb_chapter: Chapter) -> chapter.Chapter:
    """
    Convert a chapter in Jupyter Book format to the internal format
    """

    base = None
    if jb_chapter.file:
        base = section.Section(file=jb_chapter.file)

    return chapter.Chapter(
        base=base,
        sections=[jb_section.from_jupyterbook(s) for s in jb_chapter.sections],
    )


def to_jupyterbook(ic_chapter: chapter.Chapter) -> Chapter:
    """
    Convert a chapter in the internal format to Jupyter Book format
    """

    file = None
    if ic_chapter.base:
        file = ic_chapter.base.file

    return Chapter(
        file=file,
        sections=[jb_section.to_jupyterbook(s) for s in ic_chapter.sections],
    )



"""
A Part of a document, particularly a book. It is a collection
of Chapters.
"""

from pathlib import Path
from typing import Any, Callable

from pydantic import BaseModel

from icreports.tags import Taggable
from icreports.document import chapter

from .chapter import Chapter


class Part(BaseModel, Taggable):
    """
    A Part is an element of a document, particularly a book,
    that is made of a collection of Chapters.

    :param str caption: A caption or title for the Part
    :param list[Chapter] chapters: A list of the Part's Chapters
    """

    caption: str = ""
    chapters: list[Chapter] = []

    @property
    def files(self) -> list[Path]:
        """
        Paths for all content files in the Path
        """
        return [p for c in self.chapters for p in c.files]

    def has_tag(self, tag: str) -> bool:
        """
        True if any Chapter has the provided tag
        """
        return any(c.has_tag(tag) for c in self.chapters)


def filter_children(part: Part, pred: Callable[[Any], bool]) -> Part:
    """
    Return a Part with Chapters matching the predicate

    :param Part part: The part to filter
    :param Callable pred: The predicate to filter on
    """

    return Part(
        caption=part.caption,
        chapters=[chapter.filter_children(c, pred) for c in part.chapters if pred(c)],
    )


def load_content(root: Path, part: Part) -> Part:
    """
    Load content files in child Chapters
    """

    return Part(
        caption=part.caption,
        chapters=[chapter.load_content(root, c) for c in part.chapters],
    )



import os
import logging
import argparse
from pathlib import Path

from icreports.rendering import render_book

logger = logging.getLogger(__name__)


def render_book_cli(cli_args):

    source_dir = cli_args.source_dir.resolve()
    config_path = cli_args.config.resolve()
    if not config_path.exists():
        config_path = source_dir / "_config.yml"

    render_book(source_dir, cli_args.build_dir.resolve(), config_path, cli_args.rebuild)


def main_cli():

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(required=True)

    book_parser = subparsers.add_parser("book")
    book_parser.add_argument(
        "--source_dir",
        type=Path,
        default=Path(os.getcwd()),
        help="Path to the document sources",
    )
    book_parser.add_argument(
        "--config",
        type=Path,
        default=Path(os.getcwd()) / "_config.yml",
        help="Path to document config",
    )
    book_parser.add_argument(
        "--build_dir",
        type=Path,
        default=Path(os.getcwd()) / "_build",
        help="Path for build output",
    )
    book_parser.add_argument(
        "--rebuild",
        type=bool,
        default=True,
        help="If true do a clean rebuild of the book",
    )
    book_parser.set_defaults(func=render_book_cli)
    args = parser.parse_args()

    logging.basicConfig(encoding="utf-8", level=logging.INFO)

    args.func(args)


if __name__ == "__main__":

    main_cli()



import logging
from pathlib import Path

from icreports.document.hyperlink import Hyperlink

logger = logging.getLogger(__name__)


class LinkValidator:
    def __init__(self) -> None:
        pass

    def find_link(self, root: Path, link: Hyperlink, pages: list):
        resolved_link = link.resolve(root)
        for page in pages:
            if resolved_link == str(page.file_path):
                return resolved_link
        return None

    def check_for_orphaned_pages(self, links):
        orphaned_pages = []
        for file_path, count in links.items():
            if "pages" in str(file_path) and count == 0:
                orphaned_pages.append(file_path)

        if orphaned_pages:
            msg = "The following pages are not linked from anywhere:\n"
            for page in orphaned_pages:
                msg += page + "\n"
            raise RuntimeError(msg)

    def on_failed_links(self, links):
        msg = "Failed link validation. The following links don't resolve:\n"
        for href, filepath in links:
            msg += f"File: {filepath} | Link: {href}\n"
        raise RuntimeError(msg)

    def validate_links(self, pages):
        logger.info("Starting link validation")

        link_counts = {str(page.file_path): 0 for page in pages}

        failed_links = []

        for page in pages:
            page_dir = page.file_path.parent
            for link in page.links:
                href = Hyperlink(link["href"])
                if not href.is_local():
                    continue
                link_target = self.find_link(page_dir, href, pages)
                if not link_target:
                    failed_links.append((href.link, page.file_path))
                    continue
                link_counts[link_target] = link_counts[link_target] + 1

        if failed_links:
            self.on_failed_links(failed_links)

        self.check_for_orphaned_pages(link_counts)

        logger.info("Finished link validation")



"""
Module to support rendering documents to disk for a given
build context
"""

import logging
from typing import Callable
from functools import partial

from iccore import filesystem as fs

import ictasks
from ictasks import Task

from icreports.document import toc, config
from icreports.rendering import media

from .context import RenderContext, replace_dirs

logger = logging.getLogger(__name__)


def _render_build(
    contents: toc.TableOfContents,
    ctx: RenderContext,
    build: config.Build,
    copy_sources_func: Callable,
    build_funcs: dict[str, Callable],
):

    logger.info("Start rendering build: %s", build.name)

    build_dir = ctx.build_dir / build.name
    fs.clear_dir(build_dir)

    logger.info("Copying source files")
    copy_sources_func(contents, ctx, build)

    logger.info("Copying media files")
    fs.copy_files(ctx.media_dir, build_dir / ctx.config.media_dir)
    fs.copy_files(ctx.media_build_dir, build_dir / ctx.config.media_dir)

    build_ctx = replace_dirs(ctx, build_dir)
    logger.info("Rendering with %s", build_ctx)

    if "src" in build.outputs:
        fs.make_archive(
            build_ctx.build_dir / f"{ctx.config.project_name}_src",
            "zip",
            build_dir / "src",
        )

    for fmt, func in build_funcs.items():
        if fmt in build.outputs:
            logger.info("Rendering format: %s", fmt)
            func(ctx.config.project_name, build_ctx)

    logger.info("Finished rendering build: %s", build.name)


def _task_func(render_func: Callable, _: Task):
    render_func()


def render(
    contents: toc.TableOfContents,
    ctx: RenderContext,
    copy_sources_func: Callable,
    build_funcs: dict[str, Callable],
):
    """
    Render a document to disk for a given render context. User provided callbacks are
    to copy book sources to the output archive and build functions for supported
    formats.

    Particular document types, such as Book, provide suitable callbacks for each.

    :param TableOfContents contents: The content to render
    :param RenderContext ctx: The settings for the render
    :param Callable copy_sources_func: Function to copy document source files to output
    :param dict[str, Callable] build_funcs: Function to build the document per format
    """

    logging.info("Rendering document with: %s", ctx)

    media.convert(ctx.media_dir, ctx.media_build_dir, ctx.conversion_dir)

    tasks = []
    for build in ctx.config.builds:
        render_func = partial(
            _render_build, contents, ctx, build, copy_sources_func, build_funcs
        )
        tasks.append(Task(launch_func=partial(_task_func, render_func)))
    ictasks.run_funcs(tasks)



"""
Module for handling filtering of objects based on a tag
"""

from typing import NamedTuple


class Taggable:
    """
    Mixin for an entity with a tag
    """

    def has_tag(self, tag: str) -> bool:
        """
        True if the instance has the supplied tag
        """
        raise NotImplementedError()


class TagFilter(NamedTuple):
    """
    The filter to apply
    """

    includes: set[str]
    excludes: set[str]

    def __call__(self, taggable: Taggable) -> bool:
        """
        Return whether something with a tag is included in the includes list
        and not excluded from the exclides list.

        An empty includes list implies include all, unless excluded.
        """

        exclude = any(taggable.has_tag(t) for t in self.excludes)
        if self.includes:
            include = any(taggable.has_tag(t) for t in self.includes)
            return include and not exclude
        return not exclude



import logging
from pathlib import Path

from icreports.document import toc
from icreports.document.jupyter_book import toc as jb_toc
import icreports.document.config
from icreports.rendering import book, RenderContext

logger = logging.getLogger(__name__)


def render_book(source_dir: Path, build_dir: Path, config_path: Path, rebuild: bool):

    logger.info("Loading config from %s", config_path)
    config = icreports.document.config.read(config_path)

    toc_path = source_dir / "_toc.yml"
    logger.info("Reading toc from %s", toc_path)
    contents = toc.load_content(
        source_dir, jb_toc.from_jupyterbook(jb_toc.read(toc_path))
    )

    ctx = RenderContext(
        source_dir=source_dir,
        build_dir=build_dir,
        config=config,
        rebuild=rebuild,
    )
    book.render(contents, ctx)



"""
Module for the rendering context
"""

from pathlib import Path

from pydantic import BaseModel

from icreports.document.config import Config


class RenderContext(BaseModel, frozen=True):
    """
    This captures the environment and configuration settings
    needed to render a document.
    """

    source_dir: Path
    build_dir: Path
    config: Config
    rebuild: bool = True

    @property
    def media_dir(self) -> Path:
        return self.source_dir / self.config.media_dir

    @property
    def media_build_dir(self) -> Path:
        return self.build_dir / "media"

    @property
    def conversion_dir(self) -> Path:
        return self.build_dir / "conversion"


def replace_dirs(ctx: RenderContext, new_dir: Path) -> RenderContext:
    return RenderContext(
        source_dir=new_dir,
        build_dir=new_dir / "_build",
        config=ctx.config,
        rebuild=ctx.rebuild,
    )



from iccore.test_utils import get_test_data_dir, get_test_output_dir

from icreports.document.jupyter_book import toc

def test_table_of_contents():
    toc_file = get_test_data_dir() / "mock_document/_toc.yml"

    contents = toc.read(toc_file)
    contents = toc.from_jupyterbook(contents)
    contents = toc.to_jupyterbook(contents)

    toc_out = get_test_output_dir() / "_toc_test.yml"
    toc.write(toc_out, contents)

    toc_1 = toc.read(toc_out)

    toc_out.unlink()

    assert len(toc_1.parts[0].chapters[2].sections) == 2
    



"""
Module supporting processing of file content
"""

from pathlib import Path
import logging

from markdown_it import MarkdownIt
from mdit_py_plugins.front_matter import front_matter_plugin
from mdformat.renderer import MDRenderer
from bs4 import BeautifulSoup
import yaml
import pydantic

from iccore.filesystem import read_file
from icreports.hyperlink import Hyperlink

logging.getLogger("markdown_it").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)


def _md_it_wikify_link(self, tokens, idx, options, env):
    if "href" in tokens[idx].attrs:
        link = Hyperlink(link=tokens[idx].attrs["href"])
        tokens[idx].attrSet("href", link.wikify())
    return self.renderToken(tokens, idx, options, env)


def _wikify_link(token):
    if token.type != "link_open":
        return
    if "href" in token.attrs:
        link = Hyperlink(link=token.attrs["href"])
        token.attrSet("href", link.wikify())


def _visit_token(token, func):
    if token.children is not None:
        for child in token.children:
            _visit_token(child, func)
    else:
        func(token)


def wikify_links(parser, src: str) -> str:
    tokens = parser.parse(src)

    for token in tokens:
        _visit_token(token, _wikify_link)

    md_renderer = MDRenderer()
    options: dict = {}
    env: dict = {}
    return md_renderer.render(tokens, options, env)


def _parse_frontmatter(token, meta_key: str) -> list[str]:
    tags: list[str] = []
    if not token.content:
        return tags
    try:
        meta = yaml.safe_load(token.content)
    except Exception as e:
        logger.error("Failed to load frontmatter yaml")
        raise e
    if meta_key in meta and "tags" in meta[meta_key]:
        tags = meta[meta_key]["tags"]
    return tags


def _render_html(parser, src: str) -> tuple[str, list[Hyperlink]]:
    html = parser.render(src)

    soup = BeautifulSoup(html, features="html.parser")
    links = soup.find_all("a", href=True)
    return html, [Hyperlink(link=str(link)) for link in links]


class Content(pydantic.BaseModel):
    """
    This represents the content of a text file

    :param str src: The raw document source
    :param list[str] tags: Metadata tags in the document
    :param str html: The content as html
    :param list[Hyperlink]: Hyperlinks in the document
    """

    src: str = ""
    tags: list[str] = []
    html: str = ""
    links: list[Hyperlink] = []

    def has_tag(self, tag: str):
        return tag in self.tags


def load_markdown(path: Path, meta_key: str) -> Content:
    """
    Load and parse a markdown document from the provided file
    """

    src = read_file(path)

    parser = MarkdownIt().use(front_matter_plugin)
    tags = []
    for t in parser.parse(src):
        if t.type == "front_matter":
            tags = _parse_frontmatter(t, meta_key)
            break
    html, links = _render_html(parser, src)

    return Content(src=src, tags=tags, html=html, links=links)


def load(path: Path, meta_key: str = "icreports") -> Content:
    """
    Load and parse the document content from the provided file
    """

    for ext in ["md"]:
        candidate = Path(f"{path}.{ext}")
        if candidate.exists():
            if ext == "md":
                return load_markdown(candidate, meta_key)
    raise RuntimeError(f"No file with supported extension found: {path}")
